import fs from 'fs';
import path from 'path';
import fetch from 'node-fetch';
import FormData from 'form-data';
import crypto from 'crypto';
import mime from 'mime-types';
import { pureRequire } from '@resolve-js/runtime-base'; // eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore

import interopRequireDefault from '@babel/runtime/helpers/interopRequireDefault';

const createPreSignedPut = async ({
  uploaderArn,
  userId,
  encryptedUserId
}, dir) => {
  let Lambda;

  try {
    void ({
      default: Lambda
    } = interopRequireDefault(pureRequire('aws-sdk/clients/lambda')));
  } catch {}

  const lambda = new Lambda();
  const result = await (lambda === null || lambda === void 0 ? void 0 : lambda.invoke({
    FunctionName: uploaderArn,
    Payload: JSON.stringify({
      type: 'put',
      userId,
      encryptedUserId,
      dir
    })
  }).promise());
  const {
    FunctionError,
    Payload: ResponsePayload
  } = result;

  if (FunctionError != null) {
    const {
      errorMessage
    } = ResponsePayload == null ? {
      errorMessage: 'Unknown error'
    } : JSON.parse(ResponsePayload.toString());
    throw new Error(errorMessage);
  }

  return ResponsePayload != null ? JSON.parse(ResponsePayload.toString()) : null;
};

export const upload = (pool, uploadUrl, filePath) => {
  const fileSizeInBytes = fs.statSync(filePath).size;
  const fileStream = fs.createReadStream(filePath);
  const contentType = mime.contentType(path.extname(filePath)) || 'text/plain; charset=utf-8';
  return new Promise(async (resolve, reject) => {
    try {
      const res = await fetch(uploadUrl, {
        method: 'PUT',
        headers: {
          'Content-Length': fileSizeInBytes.toString(),
          'Content-Type': contentType
        },
        body: fileStream
      });

      if (!res.ok) {
        throw new Error(await res.text());
      }

      resolve();
    } catch (error) {
      reject(error);
    }
  });
};

const createPresignedPost = async ({
  uploaderArn,
  userId,
  encryptedUserId
}, dir) => {
  let Lambda;

  try {
    void ({
      default: Lambda
    } = interopRequireDefault(pureRequire('aws-sdk/clients/lambda')));
  } catch {}

  const lambda = new Lambda();
  const {
    FunctionError,
    Payload: ResponsePayload
  } = await (lambda === null || lambda === void 0 ? void 0 : lambda.invoke({
    FunctionName: uploaderArn,
    Payload: JSON.stringify({
      type: 'post',
      userId,
      encryptedUserId,
      dir
    })
  }).promise());

  if (FunctionError != null) {
    const {
      errorMessage
    } = ResponsePayload == null ? {
      errorMessage: 'Unknown error'
    } : JSON.parse(ResponsePayload.toString());
    throw new Error(errorMessage);
  }

  return ResponsePayload != null ? JSON.parse(ResponsePayload.toString()) : null;
};

export const uploadFormData = (pool, form, filePath) => {
  const fileStream = fs.createReadStream(filePath);
  const contentType = mime.contentType(path.extname(filePath)) || 'text/plain; charset=utf-8';
  form.fields.key = form.fields.Key;
  delete form.fields.Key;
  const formData = new FormData();

  for (const [key, value] of Object.entries(form)) {
    formData.append(key, value);
  }

  formData.append('file', fileStream);
  return new Promise(async (resolve, reject) => {
    try {
      const res = await fetch(form.url, {
        method: 'POST',
        headers: {
          'Content-Type': contentType
        },
        body: formData
      });

      if (!res.ok) {
        throw new Error(await res.text());
      }

      resolve();
    } catch (error) {
      reject(error);
    }
  });
};
export const createToken = ({
  encryptedUserId
}, {
  dir,
  expireTime = 3600
}) => {
  const payload = Buffer.from(JSON.stringify({
    encryptedUserId,
    dir,
    expireTime: Date.now() + expireTime * 1000
  })).toString('base64').replace(/=/g, '');
  const signature = crypto.createHmac('md5', encryptedUserId).update(payload).digest('hex');
  return `${payload}*${signature}`;
};

const createUploader = config => {
  const {
    CDN
  } = config;
  const userId = process.env['RESOLVE_USER_ID'];
  const encryptedUserId = process.env['RESOLVE_ENCRYPTED_USER_ID'];
  Object.assign(config, {
    userId,
    encryptedUserId
  });
  return Object.freeze({
    createPresignedPut: createPreSignedPut.bind(null, config),
    upload: upload.bind(null, config),
    createPresignedPost: createPresignedPost.bind(null, config),
    uploadFormData: uploadFormData.bind(null, config),
    createToken: createToken.bind(null, config),
    CDN,
    userId,
    encryptedUserId
  });
};

const getSignedPut = async (adapter, dir) => await adapter.createPresignedPut(dir);

const getSignedPost = async (adapter, dir) => await adapter.createPresignedPost(dir);

const getCDNUrl = async ({
  CDN
}) => CDN;

export const uploaderFactory = async params => {
  const {
    uploaderAdapterFactory
  } = params;

  if (uploaderAdapterFactory != null) {
    // TODO: provide support for custom uploader adapter
    const uploader = createUploader(uploaderAdapterFactory());
    process.env.RESOLVE_UPLOADER_CDN_URL = `https://${uploader.CDN}/${uploader.userId}`;
    return {
      getSignedPut: getSignedPut.bind(null, uploader),
      getSignedPost: getSignedPost.bind(null, uploader),
      getCDNUrl: getCDNUrl.bind(null, uploader),
      createToken: uploader.createToken,
      uploadPut: uploader.upload,
      uploadPost: uploader.uploadFormData
    };
  }

  return null;
};
//# sourceMappingURL=uploader-factory.js.map