"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.waitForSubscriber = exports.isRetryableServiceError = exports.eventSubscriberNotifierFactory = exports.checkError = exports.LAMBDA_TO_STEP_FUNCTION_COST_EXPENSE_THRESHOLD_MS = void 0;

var _runtimeBase = require("@resolve-js/runtime-base");

var _interopRequireDefault = _interopRequireDefault2(require("@babel/runtime/helpers/interopRequireDefault"));

function _interopRequireDefault2(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore
const LAMBDA_TO_STEP_FUNCTION_COST_EXPENSE_THRESHOLD_MS = 3000;
exports.LAMBDA_TO_STEP_FUNCTION_COST_EXPENSE_THRESHOLD_MS = LAMBDA_TO_STEP_FUNCTION_COST_EXPENSE_THRESHOLD_MS;

const waitForSubscriber = async () => await new Promise(resolve => setTimeout(resolve, 1000));

exports.waitForSubscriber = waitForSubscriber;

const checkError = (error, value) => error != null && (error.message != null && error.message.constructor === String && error.message.indexOf(`${value}`) > -1 || error.stack != null && error.stack.constructor === String && error.stack.indexOf(`${value}`) > -1 || error.name === `${value}` || error.code === `${value}`);

exports.checkError = checkError;

const isRetryableServiceError = error => checkError(error, 'TooManyRequestsException') || checkError(error, 'ServiceException'); // TODO: unnecessary destination round trip


exports.isRetryableServiceError = isRetryableServiceError;

const notifyEventSubscriber = async (runtime, destination, eventSubscriber, event) => {
  var _event$event$type;

  const log = (0, _runtimeBase.getLog)(`notifyEventSubscriber:${eventSubscriber}:${(_event$event$type = event === null || event === void 0 ? void 0 : event.event.type) !== null && _event$event$type !== void 0 ? _event$event$type : '_NO_EVENT_'}`);

  if (/^arn:aws:sqs:/.test(destination)) {
    log.debug(`sending SQS message`);
    const queueFullName = destination.split(':')[5];
    await runtime.sendSqsMessage(queueFullName, (0, _runtimeBase.createEventSubscriberNotification)(eventSubscriber, event, true));
  } else if (/^arn:aws:lambda:/.test(destination)) {
    log.debug(`invoking lambda directly`);
    const lambdaFullName = destination.split(':')[6];
    await runtime.invokeLambdaAsync(lambdaFullName, {
      resolveSource: 'BuildEventSubscriber',
      ...(0, _runtimeBase.createEventSubscriberNotification)(eventSubscriber, event, true)
    });
  } else {
    log.warn(`event subscriber destination not supported by runtime: ${destination}`);
  }
};

const eventSubscriberNotifierFactory = async params => {
  const {
    lambdaContext,
    eventSubscriberScope
  } = params;
  let getAccountIdFromLambdaContext;

  try {
    void ({
      getAccountIdFromLambdaContext
    } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/utils'));
  } catch {}

  const accountId = getAccountIdFromLambdaContext(lambdaContext);
  const {
    functionName
  } = lambdaContext;
  const region = process.env.AWS_REGION;
  const userId = process.env.RESOLVE_USER_ID;
  const functionArn = `arn:aws:lambda:${region}:${accountId}:function:${functionName}`;
  const useSqs = !!process.env.EXPERIMENTAL_SQS_TRANSPORT;

  const getNotifierLog = scope => (0, _runtimeBase.getLog)(`subscriberNotifier:${scope}`);

  const invokeLambdaAsync = async (destination, parameters) => {
    const log = getNotifierLog(`invokeLambdaAsync`);
    log.debug(`invoking lambda as event subscriber: ${destination}`);
    let invokeFunction;

    try {
      void ({
        invokeFunction
      } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda')));
    } catch {}

    await invokeFunction({
      Region: region,
      FunctionName: destination,
      Payload: parameters,
      InvocationType: 'RequestOnly',
      MaximumExecutionDuration: 200
    });
  };

  const sendSqsMessage = async (destination, parameters) => {
    const log = getNotifierLog(`sendSqsMessage`);
    log.debug(`sending SQS message to: ${destination}`);
    const queueUrl = `https://sqs.${region}.amazonaws.com/${accountId}/${destination}`;
    let sendMessage;

    try {
      void ({
        sendMessage
      } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('resolve-cloud-common/sqs')));
    } catch {}

    await sendMessage({
      Region: region,
      QueueUrl: queueUrl,
      MessageBody: JSON.stringify(parameters)
    });
  };

  const getEventSubscriberDestination = eventSubscriber => useSqs ? `arn:aws:sqs:${region}:${accountId}:${userId}-${eventSubscriberScope}-${eventSubscriber}` : functionArn;

  const invokeBuildAsync = async (parameters, timeout) => {
    if (useSqs) {
      return await sendSqsMessage(`${userId}-${eventSubscriberScope}-${parameters.eventSubscriber}`, parameters);
    }

    const lambdaEvent = {
      resolveSource: 'BuildEventSubscriber',
      ...parameters
    };

    if (timeout == null || timeout < LAMBDA_TO_STEP_FUNCTION_COST_EXPENSE_THRESHOLD_MS) {
      await new Promise(resolve => setTimeout(resolve, timeout));
      await invokeLambdaAsync(functionName, lambdaEvent);
    } else {
      let STS;

      try {
        void ({
          default: STS
        } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('aws-sdk/clients/sts')));
      } catch {}

      const {
        Arn
      } = await new STS().getCallerIdentity().promise();
      let invokeFunction;

      try {
        void ({
          invokeFunction
        } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda')));
      } catch {}

      await invokeFunction({
        Region: process.env.AWS_REGION,
        FunctionName: process.env.RESOLVE_SCHEDULER_LAMBDA_ARN,
        Payload: {
          functionName: process.env.AWS_LAMBDA_FUNCTION_NAME,
          event: lambdaEvent,
          date: new Date(Date.now() + timeout).toISOString(),
          validationRoleArn: Arn,
          principial: {
            accessKeyId: process.env.AWS_ACCESS_KEY_ID,
            secretAccessKey: process.env.AWS_SECRET_ACCESS_KEY,
            sessionToken: process.env.AWS_SESSION_TOKEN
          }
        }
      });
    }
  };

  const ensureQueue = async name => {
    if (!useSqs) {
      return;
    }

    const getTags = () => {
      const tags = {
        'resolve-deployment-id': eventSubscriberScope,
        'resolve-function-name': functionName,
        'resolve-user-id': userId
      };
      return tags;
    };

    const errors = [];
    let roleArn = null;
    let UUID = null;

    try {
      let getCallerIdentity;

      try {
        void ({
          getCallerIdentity
        } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('resolve-cloud-common/sts')));
      } catch {}

      roleArn = (await getCallerIdentity({
        Region: region
      })).Arn;
    } catch (err) {
      errors.push(err);
    }

    try {
      while (true) {
        try {
          let ensureSqsQueue;

          try {
            void ({
              ensureSqsQueue
            } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('resolve-cloud-common/sqs')));
          } catch {}

          await ensureSqsQueue({
            QueueName: `${userId}-${eventSubscriberScope}-${name}`,
            Region: region,
            Policy: {
              Version: '2008-10-17',
              Statement: [{
                Action: 'SQS:*',
                Principal: {
                  AWS: [roleArn]
                },
                Effect: 'Allow'
              }]
            },
            Tags: getTags()
          });
          break;
        } catch (error) {
          if (checkError(error, 'QueueAlreadyExists')) {
            break;
          }

          if (!(isRetryableServiceError(error) || checkError(error, 'AWS.SimpleQueueService.QueueDeletedRecently') || checkError(error, 'QueueDeletedRecently'))) {
            throw error;
          }

          await waitForSubscriber();
        }
      }
    } catch (err) {
      errors.push(err);
    }

    try {
      while (true) {
        try {
          let createEventSourceMapping;

          try {
            void ({
              createEventSourceMapping
            } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda'));
          } catch {}

          void ({
            UUID
          } = await createEventSourceMapping({
            Region: region,
            QueueName: `${userId}-${eventSubscriberScope}-${name}`,
            FunctionName: functionName,
            MaximumBatchingWindowInSeconds: 0,
            BatchSize: 10
          }));
          break;
        } catch (error) {
          if (checkError(error, 'ResourceConflictException')) {
            break;
          }

          if (!isRetryableServiceError(error)) {
            throw error;
          }

          await waitForSubscriber();
        }
      }
    } catch (err) {
      errors.push(err);
    }

    if (UUID != null) {
      try {
        while (true) {
          try {
            let getEventSourceMapping;

            try {
              void ({
                getEventSourceMapping
              } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda'));
            } catch {}

            const {
              State
            } = await getEventSourceMapping({
              Region: region,
              UUID
            });

            if (State === 'Enabled') {
              break;
            }
          } catch (error) {
            if (!isRetryableServiceError(error)) {
              throw error;
            }
          }

          await waitForSubscriber();
        }
      } catch (err) {
        errors.push(err);
      }

      try {
        let setFunctionTags;

        try {
          void ({
            setFunctionTags
          } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda'));
        } catch {}

        await setFunctionTags({
          Region: region,
          FunctionName: functionArn,
          Tags: {
            [`SQS-${eventSubscriberScope}-${name}`]: UUID
          }
        });
      } catch (err) {
        errors.push(err);
      }
    }

    if (errors.length > 0) {
      const summaryError = new Error(errors.map(({
        message
      }) => message).join('\n'));
      summaryError.stack = errors.map(({
        stack
      }) => stack).join('\n');
      throw summaryError;
    }
  };

  const deleteQueue = async name => {
    if (!useSqs) {
      return;
    }

    const errors = [];
    let functionTags = null;
    let UUID = null;
    let queueUrl = null;

    try {
      let getFunctionTags;

      try {
        void ({
          getFunctionTags
        } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda')));
      } catch {}

      functionTags = await getFunctionTags({
        Region: region,
        FunctionName: functionArn
      });
      UUID = functionTags[`SQS-${eventSubscriberScope}-${name}`];
      queueUrl = `https://sqs.${region}.amazonaws.com/${accountId}/${userId}-${eventSubscriberScope}-${name}`;
    } catch (err) {
      errors.push(err);
    }

    if (UUID != null) {
      try {
        while (true) {
          try {
            let deleteEventSourceMapping;

            try {
              void ({
                deleteEventSourceMapping
              } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda'));
            } catch {}

            await deleteEventSourceMapping({
              Region: region,
              UUID
            });
            break;
          } catch (error) {
            if (checkError(error, 'ResourceNotFoundException')) {
              break;
            }

            if (!(isRetryableServiceError(error) || checkError(error, 'ResourceInUseException'))) {
              throw error;
            }

            await waitForSubscriber();
          }
        }
      } catch (err) {
        errors.push(err);
      }

      try {
        while (true) {
          try {
            let getEventSourceMapping;

            try {
              void ({
                getEventSourceMapping
              } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/lambda'));
            } catch {}

            await getEventSourceMapping({
              Region: region,
              UUID
            });
            const error = new Error('ResourceAlreadyExists');
            error.code = 'ResourceAlreadyExists';
            throw error;
          } catch (error) {
            if (checkError(error, 'ResourceNotFoundException')) {
              break;
            }

            if (!(isRetryableServiceError(error) || checkError(error, 'ResourceAlreadyExists'))) {
              throw error;
            }
          }

          await waitForSubscriber();
        }
      } catch (err) {
        errors.push(err);
      }

      try {
        while (true) {
          try {
            let deleteSqsQueue;

            try {
              void ({
                deleteSqsQueue
              } = (0, _runtimeBase.pureRequire)('resolve-cloud-common/sqs'));
            } catch {}

            await deleteSqsQueue({
              Region: region,
              QueueName: `${userId}-${eventSubscriberScope}-${name}`,
              QueueUrl: queueUrl
            });
            break;
          } catch (error) {
            if (checkError(error, 'Failed to delete SQS queue')) {
              break;
            }

            if (!isRetryableServiceError(error)) {
              throw error;
            }

            await waitForSubscriber();
          }
        }
      } catch (err) {
        errors.push(err);
      }
    }

    if (errors.length > 0) {
      const summaryError = new Error(errors.map(({
        message
      }) => message).join('\n'));
      summaryError.stack = errors.map(({
        stack
      }) => stack).join('\n');
      throw summaryError;
    }
  };

  const notifier = async (...args) => notifyEventSubscriber({
    sendSqsMessage,
    invokeLambdaAsync
  }, ...args);

  return {
    notifyEventSubscriber: notifier,
    getEventSubscriberDestination,
    ensureQueue,
    deleteQueue,
    invokeBuildAsync
  };
};

exports.eventSubscriberNotifierFactory = eventSubscriberNotifierFactory;
//# sourceMappingURL=event-subscriber-notifier-factory.js.map