"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.uploaderFactory = exports.uploadFormData = exports.upload = exports.createToken = void 0;

var _fs = _interopRequireDefault2(require("fs"));

var _path = _interopRequireDefault2(require("path"));

var _nodeFetch = _interopRequireDefault2(require("node-fetch"));

var _formData = _interopRequireDefault2(require("form-data"));

var _crypto = _interopRequireDefault2(require("crypto"));

var _mimeTypes = _interopRequireDefault2(require("mime-types"));

var _runtimeBase = require("@resolve-js/runtime-base");

var _interopRequireDefault = _interopRequireDefault2(require("@babel/runtime/helpers/interopRequireDefault"));

function _interopRequireDefault2(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore
const createPreSignedPut = async ({
  uploaderArn,
  userId,
  encryptedUserId
}, dir) => {
  let Lambda;

  try {
    void ({
      default: Lambda
    } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('aws-sdk/clients/lambda')));
  } catch {}

  const lambda = new Lambda();
  const result = await (lambda === null || lambda === void 0 ? void 0 : lambda.invoke({
    FunctionName: uploaderArn,
    Payload: JSON.stringify({
      type: 'put',
      userId,
      encryptedUserId,
      dir
    })
  }).promise());
  const {
    FunctionError,
    Payload: ResponsePayload
  } = result;

  if (FunctionError != null) {
    const {
      errorMessage
    } = ResponsePayload == null ? {
      errorMessage: 'Unknown error'
    } : JSON.parse(ResponsePayload.toString());
    throw new Error(errorMessage);
  }

  return ResponsePayload != null ? JSON.parse(ResponsePayload.toString()) : null;
};

const upload = (pool, uploadUrl, filePath) => {
  const fileSizeInBytes = _fs.default.statSync(filePath).size;

  const fileStream = _fs.default.createReadStream(filePath);

  const contentType = _mimeTypes.default.contentType(_path.default.extname(filePath)) || 'text/plain; charset=utf-8';
  return new Promise(async (resolve, reject) => {
    try {
      const res = await (0, _nodeFetch.default)(uploadUrl, {
        method: 'PUT',
        headers: {
          'Content-Length': fileSizeInBytes.toString(),
          'Content-Type': contentType
        },
        body: fileStream
      });

      if (!res.ok) {
        throw new Error(await res.text());
      }

      resolve();
    } catch (error) {
      reject(error);
    }
  });
};

exports.upload = upload;

const createPresignedPost = async ({
  uploaderArn,
  userId,
  encryptedUserId
}, dir) => {
  let Lambda;

  try {
    void ({
      default: Lambda
    } = (0, _interopRequireDefault.default)((0, _runtimeBase.pureRequire)('aws-sdk/clients/lambda')));
  } catch {}

  const lambda = new Lambda();
  const {
    FunctionError,
    Payload: ResponsePayload
  } = await (lambda === null || lambda === void 0 ? void 0 : lambda.invoke({
    FunctionName: uploaderArn,
    Payload: JSON.stringify({
      type: 'post',
      userId,
      encryptedUserId,
      dir
    })
  }).promise());

  if (FunctionError != null) {
    const {
      errorMessage
    } = ResponsePayload == null ? {
      errorMessage: 'Unknown error'
    } : JSON.parse(ResponsePayload.toString());
    throw new Error(errorMessage);
  }

  return ResponsePayload != null ? JSON.parse(ResponsePayload.toString()) : null;
};

const uploadFormData = (pool, form, filePath) => {
  const fileStream = _fs.default.createReadStream(filePath);

  const contentType = _mimeTypes.default.contentType(_path.default.extname(filePath)) || 'text/plain; charset=utf-8';
  form.fields.key = form.fields.Key;
  delete form.fields.Key;
  const formData = new _formData.default();

  for (const [key, value] of Object.entries(form)) {
    formData.append(key, value);
  }

  formData.append('file', fileStream);
  return new Promise(async (resolve, reject) => {
    try {
      const res = await (0, _nodeFetch.default)(form.url, {
        method: 'POST',
        headers: {
          'Content-Type': contentType
        },
        body: formData
      });

      if (!res.ok) {
        throw new Error(await res.text());
      }

      resolve();
    } catch (error) {
      reject(error);
    }
  });
};

exports.uploadFormData = uploadFormData;

const createToken = ({
  encryptedUserId
}, {
  dir,
  expireTime = 3600
}) => {
  const payload = Buffer.from(JSON.stringify({
    encryptedUserId,
    dir,
    expireTime: Date.now() + expireTime * 1000
  })).toString('base64').replace(/=/g, '');

  const signature = _crypto.default.createHmac('md5', encryptedUserId).update(payload).digest('hex');

  return `${payload}*${signature}`;
};

exports.createToken = createToken;

const createUploader = config => {
  const {
    CDN
  } = config;
  const userId = process.env['RESOLVE_USER_ID'];
  const encryptedUserId = process.env['RESOLVE_ENCRYPTED_USER_ID'];
  Object.assign(config, {
    userId,
    encryptedUserId
  });
  return Object.freeze({
    createPresignedPut: createPreSignedPut.bind(null, config),
    upload: upload.bind(null, config),
    createPresignedPost: createPresignedPost.bind(null, config),
    uploadFormData: uploadFormData.bind(null, config),
    createToken: createToken.bind(null, config),
    CDN,
    userId,
    encryptedUserId
  });
};

const getSignedPut = async (adapter, dir) => await adapter.createPresignedPut(dir);

const getSignedPost = async (adapter, dir) => await adapter.createPresignedPost(dir);

const getCDNUrl = async ({
  CDN
}) => CDN;

const uploaderFactory = async params => {
  const {
    uploaderAdapterFactory
  } = params;

  if (uploaderAdapterFactory != null) {
    // TODO: provide support for custom uploader adapter
    const uploader = createUploader(uploaderAdapterFactory());
    process.env.RESOLVE_UPLOADER_CDN_URL = `https://${uploader.CDN}/${uploader.userId}`;
    return {
      getSignedPut: getSignedPut.bind(null, uploader),
      getSignedPost: getSignedPost.bind(null, uploader),
      getCDNUrl: getCDNUrl.bind(null, uploader),
      createToken: uploader.createToken,
      uploadPut: uploader.upload,
      uploadPost: uploader.uploadFormData
    };
  }

  return null;
};

exports.uploaderFactory = uploaderFactory;
//# sourceMappingURL=uploader-factory.js.map